/* This file is part of SLB
   Copyright (C) 2011 Sergey Poznyakoff

   SLB is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   SLB is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with SLB.  If not, see <http://www.gnu.org/licenses/>. */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sysexits.h>
#include <syslog.h>
#include <errno.h>
#include <signal.h>

#include <regex.h>
#include <stdbool.h>
#include <stdio.h>
#include "c-ctype.h"
#include "obstack.h"
#include "progname.h"
#include "xalloc.h"
#include <inttypes.h>
#include "inttostr.h"

#include "gettext.h"

#define _(s) gettext(s)
#define N_(s) s

#include "grecs.h"

#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>

extern int slb_sigtab[];
extern size_t slb_sigcount;
extern unsigned long slb_loop_serial;

void set_signals(void (*handler) (int signo), int *sigv, int sigc);

void config_help(void);
void config_init(void);
void config_finish(void);

extern int daemon_mode;
extern int foreground;
extern char *pidfile;
extern time_t wakeup_interval;


extern int debug_level;

#define debug(lev, s)					\
	do {						\
		if (debug_level >= (lev))		\
		       debug_printf s;			\
	} while(0)


enum node_type {
	node_unknown,
	node_variable,
	node_const,
	node_uminus,
	node_add,
	node_sub,
	node_mul,
	node_div,
	node_or,
	node_and,
	node_not,
	node_eq,
	node_ne,
	node_lt,
	node_le,
	node_gt,
	node_ge,
	node_cond,
	node_func
};

/* Maximum number of arguments to a built-in function */
#define SLB_MAX_ARGS 256

struct slb_node;
struct slb_expression;
struct slb_node_state;

typedef int (*slb_func_t) (struct slb_node_state **pstate,
			   struct slb_node *node,
			   int nargs, double *args,
			   double *res);

struct function {
	const char *name;
	slb_func_t func;
	int minargs;
	int maxargs;
};

struct slb_node {
	enum node_type n_type;
	struct slb_node *n_next;
	union {
		struct slb_node *n_arg[3];
		double n_const;
		struct slb_varref *n_var;
		struct {
			slb_func_t func;
			struct slb_node *args;
			size_t nargs;
		} n_func;
	} v;
};

struct slb_node_list {
	struct slb_node *head;
	struct slb_node *tail;
	size_t count;
};


struct symtab;

struct syment {
	char *name;
};

typedef int (*symtab_enumerator_t)(void *sym, void *data);

const char *symtab_strerror(int rc);

int symtab_lookup_or_install(void **ent,
			     struct symtab *st, void *key,
			     int *install);
void symtab_clear(struct symtab *st);
struct symtab *symtab_create(size_t elsize, 
			     unsigned (*hash_fun)(void *, unsigned long),
			     int (*cmp_fun)(const void *, const void *),
			     int (*copy_fun)(void *, void *),
			     void *(*alloc_fun)(size_t),
			     void (*free_fun)(void *));
struct symtab *symtab_create_default(size_t elsize);

void symtab_destroy(struct symtab **pst);
int symtab_remove(struct symtab *st, void *elt);
int symtab_replace(struct symtab *st, void *ent, void **old_ent);
int symtab_enumerate(struct symtab *st, symtab_enumerator_t fun, void *data);

size_t symtab_count_entries(struct symtab *st);

struct slb_varref {
	struct syment vr_sym;
	struct slb_varinstance *vr_inst;
};

#define SLV_VIF_CONST   0x01
#define SLV_VIF_USED    0x02
#define SLV_VIF_SET     0x04

struct slb_varinstance {
	struct syment vi_sym;
	const char *vi_mib;
	oid vi_oid[MAX_OID_LEN];
	size_t vi_oidlen;
	double vi_value;
	int vi_flags;
};

struct slb_expression {
	const char *ex_name;
	const char *ex_text;
	grecs_locus_t ex_locus;
	struct symtab *ex_vartab;
	struct slb_node *ex_tree;
};

struct slb_macro {
	struct syment mac_sym;
	const char *mac_exp;
};


struct slb_varinstance *varinst_lookup(struct symtab *vit,
				       const char *name, size_t len,
				       int *install);
struct slb_varinstance *varinst_lookupz(struct symtab *vit,
					const char *name, int *install);
struct slb_varref *varref_lookup(struct symtab *vit,
				 const char *name, size_t len,
				 int *install);
struct slb_varref *varref_lookupz(struct symtab *vit,
				  const char *name, int *install);
struct slb_macro *macro_lookup(struct symtab *mt,
			       const char *name, size_t len,
			       int *install);
struct slb_macro *macro_lookupz(struct symtab *mt,
				const char *name, int *install);

void varinst_clear(struct symtab *vit);
int varinst_allset(struct symtab *vit);

struct symtab *create_oidtab(struct symtab *vit);
struct slb_varinstance *oidtab_lookup(struct symtab *oidtab,
				      oid *name, size_t name_length);


int register_expression(struct slb_expression *expr);
struct slb_expression *expression_lookupz(const char *name);
struct slb_expression *expression_lookup(const char *name, size_t len);
struct slb_expression *copy_expression(struct slb_expression *src);
int fixup_expression(struct slb_expression *expr, struct symtab *vit);
int bind_expression(struct slb_expression *expr, struct symtab *vit);

int compile_expression(const char *text, grecs_locus_t *locus,
		       struct slb_expression **pexpr);
int eval_expression(struct slb_expression *expr,
		    struct slb_node_state **pstate,
		    double *pnum);
struct slb_node *copy_node(struct slb_node *node, struct symtab *vartab);

#define SLB_SRV_DISABLED 0x01
#define SLB_SRV_COMPUTED 0x02

struct slb_server {
	char *id;
	grecs_locus_t locus;
	char *peer;
	int port;
	time_t timeout;
	unsigned retries;
	char *community;
	struct symtab *varinst;
	struct symtab *oidtab; /* A reverse index to the entries from varinst */
	struct symtab *macros;
	struct slb_node_state *state;
	struct slb_expression *expr;
	struct snmp_session *sess;
	double weight;
	int flags;
	struct slb_server *next;
};

extern struct slb_server *srv_head, *srv_tail;
extern size_t srv_count;
extern struct slb_server **srvtab;
extern size_t srvtab_count;
extern unsigned snmp_skip_count;

void snmploop(void);
char *srvid(struct slb_server *srv);

extern int log_to_stderr;
extern int syslog_facility;
extern const char *syslog_tag;
extern int syslog_include_prio;

void logger_setup(void);
void logmsg(int prio, const char *fmt, ...) SLB_PRINTFLIKE(2,3);
void logmsg_at_locus(int prio, grecs_locus_t *locus, const char *fmt, ...)
	SLB_PRINTFLIKE(3,4);
void debug_printf(const char *fmt, ...) SLB_PRINTFLIKE(1,2);

int string_to_syslog_facility(const char *str, int *pfacility);


extern size_t output_tail_count;
extern size_t output_head_count;
extern char *output_file;
extern char *output_format;
extern char *begin_output_message;
extern char *end_output_message;

void slb_set_output_defaults(void);
void slb_open_output(void);
void slb_close_output(void);
void slb_check_output(void);

void slb_fmtout(void);

void collect_output_refs(void);
int bind_output_refname(struct slb_varinstance *vinst);
void undo_output_refs(void);
void report_unbound_output_refs(struct slb_server *srv);

void close_fds(fd_set *fdset);


